/*
SHG -- Spherical Harmonics Generator.
Copyright (C) 2005, 2006 Adomas Paltanavicius

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/*
  This is adapted from gsl-1.1; it's home page is
  www.gnu.org/software/gsl.
*/

#define DBL_EPSILON   (2.2204460492503131e-16)
#define LOG_DBL_MIN   (-7.0839641853226408e+02)

double
legendre_Plm (const int l, const int m, const double x)
{
  /* If l is large and m is large, then we have to worry
   * about overflow. Calculate an approximate exponent which
   * measures the normalization of this thing.
   */
  double dif = l - m;
  double sum = l + m;
  double exp_check = 0.5 * log (2.0 * l + 1.0) 
                   + 0.5 * dif * (log (dif) - 1.0)
                   - 0.5 * sum * (log (sum) - 1.0);

  if (m < 0 || l < m || x < -1.0 || x > 1.0) {
    printf ("Domain error\n");
    exit (1);
  }
  else if (exp_check < LOG_DBL_MIN + 10.0) {
    printf ("Overflow\n");
    exit (1);
  }
  else {
    /* Account for the error due to the
     * representation of 1-x.
     */
    double p_mm;     /* P_m^m(x) */
    double p_mmp1;   /* P_{m+1}^m(x) */

    /* Calculate P_m^m from the analytic result:
     *          P_m^m(x) = (-1)^m (2m-1)!! (1-x^2)^(m/2) , m > 0
     */
    p_mm = 1.0;
    if(m > 0){
      double root_factor = sqrt (1.0 - x) * sqrt (1.0 + x);
      double fact_coeff = 1.0;
      int i;
      for (i = 1; i <= m; i++) {
        p_mm *= -fact_coeff * root_factor;
        fact_coeff += 2.0;
      }
    }

    /* Calculate P_{m+1}^m. */
    p_mmp1 = x * (2 * m + 1) * p_mm;

    if (l == m) {
      return p_mm;
    }
    else if (l == m + 1) {
      return p_mmp1;
    }
    else {
      double p_ell = 0.0;
      int ell;
    
      /* Compute P_l^m, l > m + 1 by upward recurrence on l. */
      for(ell = m + 2; ell <= l; ell++) {
        p_ell = (x * (2 * ell - 1) * p_mmp1 - (ell + m - 1) * p_mm) / (ell - m);
        p_mm = p_mmp1;
        p_mmp1 = p_ell;
      }

      return p_ell;
    }
  }
}
